import { Component, OnDestroy, OnInit } from '@angular/core';
import { NbMediaBreakpointsService, NbMenuItem, NbMenuService, NbSidebarService, NbThemeService } from '@nebular/theme';

import { filter, flatMap, map, takeUntil } from 'rxjs/operators';
import { Subject, Observable, combineLatest } from 'rxjs';
import { Router } from '@angular/router';
import { AuthService } from 'app/core/services/auth.service';
import { AppStateService } from 'app/core/services/app.state.service';
import { Term } from 'app/features/modules/models/term';
import { TermService } from 'app/features/modules/services/term.service';
import { DARK_THEME } from 'app/core/styles/theme.dark';
import { SUBATO_THEME } from 'app/core/core.module';
import { SubatoPrincipal } from 'app/core/models/principal/subato-principal';
import { NotificationService } from 'app/shared/services/notification.service';
import { TopMenu } from 'app/core/services/top-menu';

@Component({
  selector: 'ngx-header',
  styleUrls: ['./header.component.scss'],
  templateUrl: './header.component.html',
})
export class HeaderComponent implements OnInit, OnDestroy {
  private destroy$: Subject<void> = new Subject<void>();
  principal: SubatoPrincipal;

  _selectedTerm: number;
  termsList: Term[] = [];

  userMenu = [{title: 'Abmelden', icon: 'log-out-outline'}];
  userPictureOnly = false;

  courseId: number | null = null;

  topMenu: NbMenuItem[] = []
  showMenu: boolean = false;

  public constructor(
    private sidebarService: NbSidebarService,
    private menuService: NbMenuService,
    private notificationService: NotificationService,
    private themeService: NbThemeService,
    public authService: AuthService,
    private termService: TermService,
    public appStateService: AppStateService,
    private breakpointService: NbMediaBreakpointsService,
    private router: Router
  ) {
   
  }

  ngOnInit() {
    this.appStateService.courseId$.subscribe(courseId => this.courseId = courseId);

    this.subscribeTopMenu();

    this.authService.principal$.pipe(
      takeUntil(this.destroy$),
    ).subscribe(principal => {
      this.principal = principal;
    });

    this.appStateService.term$
      .pipe(takeUntil(this.destroy$))
      .subscribe(term => {
        if (term == null) return;
        this._selectedTerm = term.id;
      });

    const { lg } = this.breakpointService.getBreakpointsMap();
    this.themeService.onMediaQueryChange()
      .pipe(
        map(([, currentBreakpoint]) => currentBreakpoint.width < lg),
        takeUntil(this.destroy$),
      )
      .subscribe((isLessThan: boolean) => {
        this.userPictureOnly = isLessThan;
        this.showMenu = !isLessThan;
      });

    this.termService.getAll()
      .pipe(takeUntil(this.destroy$))
      .subscribe(terms => {
        this.termsList = terms;
      });
    
    this.menuService.onItemClick()
      .pipe(
        filter(({ tag }) => tag === 'user-menu'),
        map(({ item: { title } }) => title),
        flatMap(() => this.authService.logout())
      )
      .subscribe(() => { 
        this.navigateHome();
      }, err => {
        this.notificationService.showError("Fehler beim Logout");
      });
  }

  ngOnDestroy() {
    this.destroy$.next();
    this.destroy$.complete();
  }

  toggleTheme() {
    this.appStateService.setDarkMode(!this.appStateService.darkMode);
  }

  toggleSidebar(): boolean {
    this.sidebarService.toggle(true, 'menu-sidebar');
    return false;
  }

  navigateHome() {
    this.router.navigate(this.principal ? ["/dashboard"] : ["/"]);
    //this.menuService.navigateHome();
    return false;
  }

  login() {
    this.authService.login();
  }

  get selectedTerm(): number {
    return this._selectedTerm;
  }

  set selectedTerm(value: number) {
    this._selectedTerm = value;
    this.appStateService.setTerm(this.termsList.find(t => t.id === value), true);
    if(this.router.url !== '/modules' && this.router.url !== '/submissions' && this.router.url !== '/appointments') {
      //this.navigateHome();
    }
  }

  subscribeTopMenu() {
    combineLatest(this.authService.principal$, this.appStateService.courseId$)
      .pipe(takeUntil(this.destroy$))
      .subscribe(([principal, courseId]) => {
        this.topMenu = TopMenu.getMenu(this.appStateService.currentUrl, courseId, principal);
    });
  }
}
