import { Component, EventEmitter, Input, OnInit, Output } from '@angular/core';
import { TaskService } from 'app/features/tasks/services/task.service';
import { BaseComponent } from 'app/shared/components/base/base.component';
import { BackendError } from 'app/shared/models/backend-error';
import { DataQueryParams } from 'app/shared/models/data-query-params';
import { Page } from 'app/shared/models/page';
import { combineLatest, Subject } from 'rxjs';
import { takeUntil, debounceTime } from 'rxjs/operators';
import { Task } from 'app/features/tasks/models/task';
import { TaskPool } from 'app/features/tasks/models/task-pool';
import { TaskPoolService } from 'app/features/tasks/services/task-pool.service';

@Component({
  selector: 'ngx-exercise-task-selection',
  templateUrl: './exercise-task-selection.component.html',
  styleUrls: ['./exercise-task-selection.component.scss']
})
export class ExerciseTaskSelectionComponent extends BaseComponent implements OnInit {
  @Input() hiddenTasks: Task[] = [];
  @Output() taskSelected: EventEmitter<Task> = new EventEmitter<Task>();

  loading: boolean = true;
  error: BackendError;

  query: string;
  private queryChanged: Subject<string> = new Subject<string>();
  tasks: Task[] = [];

  page: Page<Task> | null;
  loadingPage = false;

  pools: TaskPool[] = [];
  selectedPool: TaskPool | null = null;

  constructor(
    private taskService: TaskService,
    private taskPoolService: TaskPoolService
  ) {
    super();
  }

  ngOnInit(): void {
    combineLatest(this.taskService.getAll(), this.taskPoolService.getAll())
    .pipe(
      takeUntil(this.unsubscribe)
    ).subscribe(([tasks, pools]) => {
      this.tasks = tasks.content;
      this.pools = pools;

      this.page = tasks;
      
      this.loading = false;
    }, (error: any) => {
      if (BackendError.isBackendError(error)) {
        this.error = error;
        this.loading = false;
      } else {
        throw error;
      }
    });

    this.queryChanged
      .pipe(
        takeUntil(this.unsubscribe),
        debounceTime(500),
      )
      .subscribe(query => {
        this.query = query;
        this.page = null;
        this.loadNextTasks(true);
      });
  }

  selectPool(pool: TaskPool) {
    this.selectedPool = pool;
    this.page = null;
    this.loadNextTasks(true);
  }

  loadNextTasks(reset: boolean = false) {
    if (this.loadingPage || (this.page != null && !this.page.hasNext())) { return }
    this.loadingPage = true;
    
    let projection = DataQueryParams.next(this.page);
    this.taskService.getAll({ q: this.query, poolId: this.selectedPool?.id, projection: projection })
      .subscribe(page => {
        this.page = page;
        if(reset) {
          this.tasks = [];
        }
        this.tasks.push(...page.content);
        this.loadingPage = false;
      }, error => {
        if (BackendError.isBackendError(error)) {
          this.error = error;
          this.loading = false;
        } else {
          throw error;
        }
      });
  }

  get selectableTasks() {
    return this.tasks.filter(m => !this.isTaskHidden(m))
  }

  isTaskHidden(task: Task) {
    return this.hiddenTasks.map(m => m.id).indexOf(task.id) != -1;
  }

  filterTasks(query) {
    this.queryChanged.next(query);
  }

  selectTask(task: Task) {
    this.taskSelected.emit(task);
  }
}
