import { AfterViewInit, Component, ElementRef, OnInit, QueryList, ViewChild, ViewChildren } from '@angular/core';
import { ActivatedRoute, Router } from '@angular/router';
import { AuthService } from 'app/core/services/auth.service';
import { BaseComponent } from 'app/shared/components/base/base.component';
import { NotificationService } from 'app/shared/services/notification.service';
import { Task } from 'app/features/tasks/models/task';
import { Submission } from '../../models/submission';
import { SolutionService } from '../../../solutions/services/solution.service';
import { SubmissionService } from '../../services/submission.service';
import { combineLatest, forkJoin, Observable, of } from 'rxjs';
import { takeUntil, debounceTime, flatMap, map, filter } from 'rxjs/operators';
import { TaskInstance } from '../../models/task-instance';
import { UntypedFormBuilder, UntypedFormGroup, Validators } from '@angular/forms';
import { NbDialogService, NbTabsetComponent } from '@nebular/theme';
import { DialogSubmissionDecisionComponent } from '../../dialogs/dialog-submission-decision/dialog-submission-decision.component';
import { ScrollHelper } from 'app/shared/helper/scroll.helper';
import { BackendError } from 'app/shared/models/backend-error';
import { ExerciseService } from '../../services/exercise.service';
import { SubmissionsProgress } from '../../models/submissions-progress';
import { ErrorContainerConfig } from 'app/shared/components/error-container/error-container.component';
import { AppStateService } from 'app/core/services/app.state.service';
import { Breadcrumb } from 'app/shared/components/breadcrumbs/breadcrumbs.component';
import { SubmissionSelectionComponent } from '../../components/submission-selection/submission-selection.component';
import { SubatoPrincipal } from 'app/core/models/principal/subato-principal';
import { CodemirrorComponent } from '@ctrl/ngx-codemirror';
import { Solution } from 'app/features/solutions/models/solution';


@Component({
  selector: 'ngx-submission',
  templateUrl: './submission.component.html',
  styleUrls: ['./submission.component.scss']
})
export class SubmissionComponent extends BaseComponent implements OnInit, AfterViewInit {
  submissionSelection: SubmissionSelectionComponent;
  @ViewChildren(SubmissionSelectionComponent) submissionSelectionQuery: QueryList<SubmissionSelectionComponent>;
  @ViewChild("solutionTabset") solutionTabset: NbTabsetComponent;
  @ViewChildren('taskInstanceSummary', { read: ElementRef }) taskInstanceSummary: QueryList<ElementRef>;
  @ViewChildren(CodemirrorComponent) codemirrorComponents: QueryList<CodemirrorComponent>;

  submission: Submission;
  progress: SubmissionsProgress;
  loading: boolean;
  loadingTask: boolean;
  pointsForm: UntypedFormGroup;
  taskId: number;
  courseId: number;
  unitId: number;
  principal: SubatoPrincipal;
  saving: boolean = false;

  hasValidSolution: boolean;
  currentTaskInstance: TaskInstance;
  currentTask: Task;
  currentSolution: Solution;
  taskInstances: TaskInstance[];
  prevTaskInstance: TaskInstance | null = null;
  nextTaskInstance: TaskInstance | null = null;
  nextSubmission: Submission | null = null;
  submissions: Submission[] = [];
  historySolutions: Solution[] = [];

  error: BackendError;
  errorConfig: ErrorContainerConfig = {
    notFound: {
      message: 'Diese Abgabe konnten wir leider nicht finden.'
    }
  };

  breadcrumbs: Breadcrumb[] = [];

  showFinished = false;

  constructor(
    private solutionService: SolutionService,
    private appStateService: AppStateService,
    private exerciseService: ExerciseService,
    private submissionService: SubmissionService,
    private authService: AuthService,
    private notificationService: NotificationService,
    private route: ActivatedRoute,
    private router: Router,
    private dialogService: NbDialogService,
    private fb: UntypedFormBuilder) {
    super();
  }

  buildBreadcrumbs(submission: Submission) {
    this.breadcrumbs = [
      {
        id: 'exercises',
        params: {
          course_id: submission.exercise.course.id
        },
      },
      {
        id: 'exercise',
        params: {
          exercise_id: submission.exercise.id
        },
        ctx: submission.exercise
      },
      {
        id: 'submissions',
        params: {
          exercise_id: submission.exercise.id
        },
      },
      {
        id: 'submission',
        params: {
          submission_id: submission.id
        },
        ctx: submission
      },
    ];
  }

  ngAfterViewInit() {
    this.submissionSelectionQuery.changes.subscribe((comps: QueryList<SubmissionSelectionComponent>) => {
      this.submissionSelection = comps.first;
    });
  }

  get isLecturerOrAdmin() {
    return this.principal.isLecturerOrAdmin(this.courseId);
  }

  ngOnInit(): void {
    this.loading = true;
    combineLatest(this.route.paramMap, this.route.queryParamMap, this.authService.principal$)
      .pipe(
        // https://github.com/angular/angular/issues/26764
        debounceTime(100),
        takeUntil(this.unsubscribe),
        flatMap(([params, queryParams, principal]) => {
          this.principal = principal;

          let submissionId = Number(params.get("submission_id"));
          this.courseId = params.has("course_id") ? Number(params.get("course_id")) : null;
          this.taskId = queryParams.has("taskId") ? Number(queryParams.get("taskId")) : null;
          this.unitId = queryParams.has("unitId") ? Number(queryParams.get("unitId")) : null;
          return this.submissionService.getById(submissionId);
        }),
        flatMap(submission => {
          this.submission = submission;
          this.updateNextSubmission();
          this.appStateService.setCourse(submission.exercise.course.id);
          this.buildBreadcrumbs(submission);

          var loadProgress = this.unitId ? this.submissionService.getSubmissionProgress(this.submission.exercise.id, this.unitId) : of([]);
          return loadProgress;
        })
      )
      .subscribe(unitProgresses => {
        this.progress = unitProgresses[0]?.progress;
        this.loading = false;

        if (this.taskInstances != null) {
          this.selectTask(this.taskInstances[0]);
        }
      }, (error: any) => {
        if (BackendError.isBackendError(error)) {
          this.error = error;
          this.loading = false;
        } else {
          throw error;
        }
      });
  }

  loadSolution(id: number): Observable<Solution> {
    return this.solutionService.getById(id).pipe(
      takeUntil(this.unsubscribe)
    );
  }

  get accent() {
    if (this.submission.passed == null) {
      return 'basic';
    }

    return this.submission.passed ? 'success' : 'danger';
  }

  get showHistoryDot() {
    return this.historySolutions.length != 0 && 
    (this.historySolutions.some(s => s.commentsCount != 0) || this.historySolutions.some(s => s.delayed || s.attemptsExceeded));
  }

  initForTask(taskInstance: TaskInstance) {
    this.scrollBackToTop();

    this.showFinished = false;
    this.currentTaskInstance = taskInstance;
    this.updatePrevAndNext();

    if (taskInstance == null) {
      return of(null);
    }

    if (taskInstance.reachablePoints != null) {
      this.createPointsForm(taskInstance);
    } else {
      this.pointsForm = null;
    }

    return forkJoin(
      this.exerciseService.getTaskInstance(taskInstance.exercise.id, taskInstance.task.id),
      this.solutionService.getSolutions({
        taskId: taskInstance.task.id,
        submissionId: this.submission.id,
        valid: true
      }, { sortBy: 'submittedAt', sortOrder: 'desc', pageSize: 1 })
    ).pipe(
      takeUntil(this.unsubscribe),
      flatMap(([taskInstance, solutions]) => {
        this.currentTask = taskInstance.task;
        let solution = solutions.content[0];
        this.hasValidSolution = solution != null;
        let loadSolution = solution ? this.loadSolution(solution.id) : of<Solution>(null);
        return loadSolution;
      }),
      map(solution => {
        this.currentSolution = solution;
        this.refreshEditors();
      })
    );
  }

  onTasksLoaded(taskInstances: TaskInstance[]) {
    this.taskInstances = taskInstances;

    var ti = this.taskId
      ? taskInstances.find(ti => ti.task.id == this.taskId)
      : taskInstances[0];

    this.selectTask(ti);
  }

  onSubmissionsLoaded(submissions: Submission[]) {
    this.submissions = submissions;
    this.updateNextSubmission();
  }

  onSolutionHistoryLoaded(solutions: Solution[]) {
    this.historySolutions = solutions;
  }

  selectTask(taskInstance: TaskInstance) {
    this.loadingTask = true;
    this.initForTask(taskInstance)
      .subscribe(() => {
        this.loadingTask = false;
      }, err => {
        if (BackendError.isBackendError(err)) {
          this.notificationService.showError("Fehler beim Laden");
        } else {
          throw err;
        }
      });
  }

  refreshEditors() {
    this.codemirrorComponents.forEach(component => {
      if (component?.codeMirror) {
        component.codeMirror.refresh();
      }
    });
  }

  selectSolutionTab() {
    let tab = this.solutionTabset?.tabs?.find(t => t.tabId == 'solution');
    if(tab) {
      this.solutionTabset.selectTab(tab);
    }
  }

  onSolutionSelected(solution: Solution) {
    this.selectSolutionTab();

    this.loadSolution(solution.id)
      .subscribe(solution => {
        this.currentSolution = solution;
        this.refreshEditors();
      });
  }

  createPointsForm(taskInstance: TaskInstance) {
    var points = this.submission.pointRatings.find(r => r.taskInstance.task.id == taskInstance.task.id)?.points ?? '';

    this.pointsForm = this.fb.group({
      points: [points, [Validators.pattern("^(-[0-9]+)|([0-9]+)$")]],
    });
  }

  updatePrevAndNext() {
    this.prevTaskInstance = this.currentTaskInstance.getPrevTaskInstance(this.taskInstances);
    this.nextTaskInstance = this.currentTaskInstance.getNextTaskInstance(this.taskInstances);
  }

  updateNextSubmission() {
    this.nextSubmission = this.submission.getNextSubmission(this.submissions);
  }

  prev() {
    let ti = this.prevTaskInstance;
    if (ti) {
      this.selectTask(ti);
    }
  }


  pointsChanged(): boolean {
    let points = parseInt(this.pointsForm.get('points').value);
    let currentPoints = this.submission.pointRatings.find(
      r => r.taskInstance.task.id == this.currentTaskInstance.task.id
    )?.points;

    return points != currentPoints;
  }

  savePoints(): Observable<void> {
    if (this.pointsForm && this.isLecturerOrAdmin && this.pointsChanged()) {
      let points = parseInt(this.pointsForm.get('points').value);

      return this.submissionService.updatePoints(this.submission.id, {
        points: points,
        taskId: this.currentTaskInstance.task.id
      }).pipe(
        takeUntil(this.unsubscribe),
        map(submission => {
          this.submission.totalPoints = submission.totalPoints;
          let submissionInList = this.submissions.find(s => s.id == this.submission.id);
          if(submissionInList) {
            submissionInList.totalPoints = submission.totalPoints;
          }
          this.submission.pointRatings = submission.pointRatings;
          if (points < 0 || points > this.currentTaskInstance.reachablePoints) {
            this.notificationService.showWarning("Punktzahl gespeichert aber hast du sie richtig eingegeben?");
          } else {
            this.notificationService.showSuccess("Punktzahl gespeichert");
          }
        })
      );
    } else {
      return of(null);
    }
  }

  private scrollBackToTop() {
    let element = this.taskInstanceSummary.first?.nativeElement;
    if (element) {
      ScrollHelper.toElementWithOffset(element, -200);
    }
  }

  next() {
    this.saving = true;

    this.savePoints().subscribe(() => {
      this.saving = false;

      let ti = this.nextTaskInstance;
      if (ti) {
        this.selectTask(ti);
      }
    }, (error) => {
      this.saving = false;

      if (BackendError.isBackendError(error)) {
        this.notificationService.showError("Fehler beim Speichern");
      } else {
        throw error;
      }
    }, () => this.saving = false);
  }

  makeDecision(): Observable<void> {
    return this.dialogService.open(DialogSubmissionDecisionComponent, {
      context: {
        submission: this.submission
      }
    }).onClose.pipe(
      takeUntil(this.unsubscribe),
      filter(result => result && !result.cancel),
      map((result: any) => {
        if (result) {
          this.setPassed(result.passed);
        }
      })
    );
  }

  openDecision() {
    this.makeDecision().subscribe(() => { });
  }

  selectSubmission(submission: Submission) {
    this.router.navigate(['submissions', submission.id], {
      queryParams: {
        unitId: this.unitId,
        //exerciseId: this.exerciseId
      }
    });
  }

  navigateToNextSubmission() {
    this.saving = true;

    this.savePoints().pipe(
      flatMap(() => this.makeDecision()),
    ).subscribe(() => {
      this.saving = false;
      this.selectSubmission(this.nextSubmission);
    }, (error) => {
      this.saving = false;

      if (BackendError.isBackendError(error)) {
        this.notificationService.showError("Fehler beim Speichern");
      } else {
        throw error;
      }
    }, () => this.saving = false)
  }

  finish() {
    this.saving = true;

    this.savePoints().pipe(
      flatMap(() => this.makeDecision())
    ).subscribe(() => {
      this.showFinished = true;
      this.currentTaskInstance = null;
      this.saving = false;
    }, (error) => {
      this.saving = false;

      if (BackendError.isBackendError(error)) {
        this.notificationService.showError("Fehler beim Speichern");
      } else {
        throw error;
      }
    }, () => this.saving = false)
  }

  setPassed(passed?: boolean) {
    this.saving = true;

    this.submissionService.updatePassed(this.submission.id, { passed: passed })
      .pipe(takeUntil(this.unsubscribe))
      .subscribe(submission => {
        this.submission.checkedBy = submission.checkedBy;
        this.submission.passed = submission.passed;
        this.submissionSelection?.update(submission);
        this.progress = this.progress?.update(this.submissions);
        this.saving = false;

        this.notificationService.showSuccess("Bewertung erfolgreich gespeichert");
      }, (error) => {
        this.saving = false;

        if (BackendError.isBackendError(error)) {
          this.notificationService.showError("Fehler beim Speichern");
        } else {
          throw error;
        }
      }, () => this.saving = false);
  }
}
