import { Component, Input, OnInit } from '@angular/core';
import { UntypedFormBuilder, UntypedFormGroup, Validators } from '@angular/forms';
import { Router, ActivatedRoute } from '@angular/router';
import { FormComponent } from 'app/shared/components/base/form.component';
import { Iso8601Helper } from 'app/shared/helper/iso8601.helper';
import { BackendError } from 'app/shared/models/backend-error';
import { NotificationService } from 'app/shared/services/notification.service';
import { takeUntil, flatMap } from 'rxjs/operators';
import { CreateTermDto } from '../../dto/create-term.dto';
import { Term } from '../../models/term';
import { TermService } from '../../services/term.service';

@Component({
  selector: 'ngx-term-form',
  templateUrl: './term-form.component.html',
  styleUrls: ['./term-form.component.scss']
})
export class TermFormComponent extends FormComponent implements OnInit {
  loading: boolean = true;
  saving: boolean = false;
  termForm: UntypedFormGroup;
  @Input() term: Term;

  constructor(
    private termService: TermService,
    private router: Router,
    private route: ActivatedRoute,
    private notificationService: NotificationService,
    private fb: UntypedFormBuilder) {
    super();
  }

  createForm(term?: Term) {
    this.termForm = this.fb.group({
      name: [term?.name ?? '', Validators.required],
      lecturesStartAt: [term?.lecturesStartAt ?? ''],
      lecturesEndAt: [term?.lecturesEndAt ?? ''],
    });
  }

  ngOnInit(): void {
    this.createForm(this.term);
  }

  cancel() {
    this.router.navigate(['/', 'term']);
  }

  save() {
    this.saving = true;
    let dto: CreateTermDto = {
      name: this.termForm.get('name').value,
      lecturesStartAt: Iso8601Helper.toString(this.termForm.get('lecturesStartAt').value, true),
      lecturesEndAt: Iso8601Helper.toString(this.termForm.get('lecturesEndAt').value, true),
    }

    if (this.term != null) {
      this.termService.update(this.term.id, dto)
        .pipe(takeUntil(this.unsubscribe))
        .subscribe(term => {
          this.saving = false;
          this.notificationService.showSuccess("Semester erfolgreich aktualisiert");
          this.router.navigate(['/', 'term']);
        }, (error: any) => {
          this.saving = false;
          if (BackendError.isBackendError(error)) {
            if (error.isValidation()) {
              this.errors = this.getValidationErrors(error);
            } 
  
            this.notificationService.showError(
              error.format(
                "Fehler bei Aktualisierung des Semesters",
              )
            );
          } else {
            throw error;
          }
        });
    } else {
      this.termService.create(dto)
        .pipe(takeUntil(this.unsubscribe))
        .subscribe(term => {
          this.saving = false;
          this.notificationService.showSuccess("Semester erfolgreich erstellt");
          this.router.navigate(['/', 'term']);
        }, (error: any) => {
          this.saving = false;
          if (BackendError.isBackendError(error)) {
            this.notificationService.showError(
              error.format(
                "Fehler bei Erstellung des Semesters" 
              )
            );
          } else {
            throw error;
          }
        });
    }
  }
}
