package de.hsrm.sls.subato.aor.term.imp.unit;

import de.hsrm.sls.subato.aor.term.model.ModulePartInstance;
import de.hsrm.sls.subato.aor.term.model.ModulePartInstanceEvent;
import de.hsrm.sls.subato.courses.event.CourseUnitEvent;
import de.hsrm.sls.subato.courses.event.CourseUnitEventDefinition;
import de.hsrm.sls.subato.courses.unit.CourseUnit;
import de.hsrm.sls.subato.courses.unit.CourseUnitRepository;
import de.hsrm.sls.subato.courses.unit.Weekday;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

import java.time.LocalDate;
import java.time.LocalDateTime;
import java.util.ArrayList;
import java.util.List;
import java.util.stream.Collectors;

@Service
@Transactional
public class CourseUnitEventImportService {
    @Autowired
    CourseUnitRepository courseUnitRepository;

    public void importEvents(CourseUnit courseUnit, ModulePartInstance mpi) {
        if(mpi.getEvents().size() == 0) {
            return;
        }

        courseUnit.getEventDefinitions().clear();

        var allUnitEvents = new ArrayList<CourseUnitEvent>();
        for(var mpiEvent : mpi.getEvents()) {
            var events = createEvents(courseUnit, mpiEvent);
            var eventDefinition = createEventDefinition(courseUnit, mpiEvent, events);
            events.forEach(e -> e.setEventDefinition(eventDefinition));

            allUnitEvents.addAll(events);
            courseUnit.getEventDefinitions().add(eventDefinition);
        }

        removeMissingEvents(courseUnit, allUnitEvents);
        courseUnitRepository.save(courseUnit);
    }

    private CourseUnitEventDefinition createEventDefinition(CourseUnit courseUnit, ModulePartInstanceEvent mpiEvent, List<CourseUnitEvent> events) {
        var eventDefinition = new CourseUnitEventDefinition();
        eventDefinition.setUnit(courseUnit);
        eventDefinition.setRoom(mpiEvent.getRoom());
        eventDefinition.setTime(mpiEvent.getStart());

        updateWeekday(eventDefinition, events);
        return eventDefinition;
    }

    private void updateWeekday(CourseUnitEventDefinition eventDefinition, List<CourseUnitEvent> evts) {
        if(evts.size() > 0) {
            var weekdayIdx = evts.stream().findFirst().get().getTakesPlaceAt().getDayOfWeek().getValue() - 1;
            var weekday = Weekday.values()[weekdayIdx];
            eventDefinition.setWeekday(weekday);
        }
    }

    private List<CourseUnitEvent> createEvents(CourseUnit courseUnit,  ModulePartInstanceEvent mpiEvent) {
        List<CourseUnitEvent> events = new ArrayList<>();

        for(var termin : mpiEvent.getTermine()) {
            var date = LocalDate.parse(termin);
            var takesPlaceAt = LocalDateTime.of(date, mpiEvent.getStart());

            CourseUnitEvent event = getOrCreateEvent(courseUnit, takesPlaceAt);
            events.add(event);
        }

        return events;
    }

    private CourseUnitEvent getOrCreateEvent(CourseUnit courseUnit, LocalDateTime takesPlaceAt) {
        var event = courseUnit.getEvents().stream().filter(e -> e.getTakesPlaceAt().equals(takesPlaceAt)).findFirst().orElse(null);
        if(event == null) {
            event = new CourseUnitEvent();
            event.setUnit(courseUnit);
            event.setTakesPlaceAt(takesPlaceAt);
            courseUnit.getEvents().add(event);
        }
        return event;
    }

    private void removeMissingEvents(CourseUnit courseUnit, ArrayList<CourseUnitEvent> events) {
        var removeEvents = courseUnit.getEvents().stream()
                .filter(e -> !events.contains(e))
                .collect(Collectors.toList());
        courseUnit.getEvents().removeAll(removeEvents);
    }
}
