package de.hsrm.sls.subato.courses.unit;

import de.hsrm.sls.subato.auth.principal.PrincipalService;
import de.hsrm.sls.subato.courses.course.Course;
import de.hsrm.sls.subato.courses.course.CourseRepository;
import de.hsrm.sls.subato.courses.event.CourseUnitEvent;
import de.hsrm.sls.subato.courses.membership.CourseStaffMembership;
import de.hsrm.sls.subato.courses.membership.CourseStaffMembershipService;
import de.hsrm.sls.subato.courses.membership.StaffGroup;
import de.hsrm.sls.subato.user.User;
import de.hsrm.sls.subato.user.UserFilter;
import de.hsrm.sls.subato.user.UserService;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.domain.PageRequest;
import org.springframework.data.domain.Pageable;
import org.springframework.stereotype.Service;

import java.util.ArrayList;
import java.util.Comparator;
import java.util.List;
import java.util.stream.Collectors;

@Service
public class CourseUnitService {
    @Autowired
    private CourseUnitRepository courseUnitRepository;

    @Autowired
    private PrincipalService principalService;

    @Autowired
    private UserService userService;

    @Autowired
    private CourseStaffMembershipService staffMembershipService;

    public CourseUnit findById(long id) {
        return findById(id, true);
    }

    public CourseUnit findById(long id, boolean raiseException) {
        var course = courseUnitRepository.findById(id);

        if(raiseException) {
            return course.orElseThrow(() -> new CourseUnitNotFoundException(id));
        } else {
            return course.orElse(null);
        }
    }

    private void applySort(List<CourseUnit> units) {
        units.sort(Comparator.comparing((CourseUnit o) -> o.getType().getOrder()).thenComparing(CourseUnit::getName));
    }

    public List<CourseUnit> findByCourse(Course course) {
        var units = courseUnitRepository.findByCourseId(course.getId());
        this.applySort(units);
        return units;
    }

    public List<CourseUnit> getParticipatingUnits(Course course, User user) {
        var units = courseUnitRepository.findParticipatingUnits(course.getId(), user.getId());
        this.applySort(units);
        return units;
    }

    public List<CourseUnit> getLecturingUnits(User user, Course course) {
        var units = course.getUnits().stream()
                .filter(unit -> unit.getLecturers().stream().anyMatch(l -> l.getId() == user.getId()))
                .distinct()
                .collect(Collectors.toList());
        this.applySort(units);
        return units;
    }

    List<CourseUnitEvent> getEvents(CourseUnit unit) {
        var events = unit.getEvents();
        events.sort(Comparator.comparing(CourseUnitEvent::getTakesPlaceAt));
        return events;
    }

    List<User> getParticipants(CourseUnit unit) {
        return userService.findAll(new UserFilter().unitId(unit.getId()), userService.nameSort());
    }

    public boolean isParticipant(CourseUnit unit, User user) {
        return unit.getParticipants().stream().filter(m -> m.getId() == user.getId()).count() != 0;
    }

    public void addParticipant(CourseUnit unit, User user) {
        var isParticipant = unit.getParticipants().stream()
                .filter(u -> u.getId() == user.getId())
                .count() != 0;

        if(isParticipant) {
            throw new UserIsAlreadyParticipantException();
        }

        unit.getParticipants().add(user);

        unit = courseUnitRepository.save(unit);
        principalService.evictPrincipal(user);
    }

    public void removeParticipant(CourseUnit unit, User user) {
        var participant = unit.getParticipants().stream()
                .filter(u -> u.getId() == user.getId())
                .findFirst().orElse(null);

        if(participant == null) {
            throw new UserIsNotAParticipantException();
        }

        unit.getParticipants().remove(participant);
        unit = courseUnitRepository.save(unit);
        principalService.evictPrincipal(participant);
    }

    public CourseUnit create(CreateCourseUnitRequest request) {
        var unit = new CourseUnit();
        unit.setAorId(request.getAorId());
        unit.setCourse(request.getCourse());
        unit.setName(request.getName());
        unit.setType(request.getType());
        unit.setLecturers(request.getLecturers());
        unit.setParticipants(request.getParticipants());

        switch(unit.getType()) {
            case Lecture -> addMembershipIfMissing(request.getCourse(), unit.getLecturers(), StaffGroup.ADMIN);
            case Lab -> addMembershipIfMissing(request.getCourse(), unit.getLecturers(), StaffGroup.LECTURER);
            case Tut -> addMembershipIfMissing(request.getCourse(), unit.getLecturers(), StaffGroup.TUTOR);
        }

        unit = courseUnitRepository.save(unit);
        request.getCourse().getUnits().add(unit);
        return unit;
    }

    private void addMembershipIfMissing(Course course, List<User> users, StaffGroup group) {
        for (var user : users) {
            if(!staffMembershipService.hasMembership(course, user, group)) {
                staffMembershipService.addMembership(course, user, group);
            }
        }
    }
}
