package de.hsrm.sls.subato.exercises.exercise;

import de.hsrm.sls.subato.auth.principal.PrincipalService;
import de.hsrm.sls.subato.auth.policy.CheckPolicyRequest;
import de.hsrm.sls.subato.auth.policy.PolicyEvaluator;
import de.hsrm.sls.subato.auth.policy.courses.CoursesAction;
import de.hsrm.sls.subato.auth.policy.exercises.ExercisesAction;
import de.hsrm.sls.subato.courses.course.CourseNotFoundException;
import de.hsrm.sls.subato.courses.course.CourseService;
import de.hsrm.sls.subato.exercises.exercise.progress.ExerciseProgressDto;
import de.hsrm.sls.subato.exercises.exercise.progress.ExerciseProgressResultDto;
import de.hsrm.sls.subato.exercises.exercise.progress.ExerciseProgressResultMapper;
import de.hsrm.sls.subato.exercises.exercise.progress.ExerciseProgressService;
import de.hsrm.sls.subato.exercises.taskinstance.*;
import de.hsrm.sls.subato.shared.DangerousActionException;
import de.hsrm.sls.subato.shared.FileResponseBuilder;
import de.hsrm.sls.subato.shared.helper.ResourceHelper;
import de.hsrm.sls.subato.shared.helper.ZipHelper;
import de.hsrm.sls.subato.tasks.stef.file.FileRefDto;
import de.hsrm.sls.subato.tasks.stef.file.FileRefMapper;
import de.hsrm.sls.subato.tasks.task.AssetNotFoundException;
import de.hsrm.sls.subato.tasks.task.FileNotFoundException;
import de.hsrm.sls.subato.tasks.task.TaskNotFoundException;
import de.hsrm.sls.subato.tasks.task.TaskService;
import io.swagger.v3.oas.annotations.Operation;
import io.swagger.v3.oas.annotations.Parameter;
import io.swagger.v3.oas.annotations.responses.ApiResponse;
import io.swagger.v3.oas.annotations.responses.ApiResponses;
import io.swagger.v3.oas.annotations.tags.Tag;
import org.apache.commons.lang3.tuple.Pair;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.core.io.Resource;
import org.springframework.core.io.UrlResource;
import org.springframework.http.ResponseEntity;
import org.springframework.web.bind.annotation.*;
import org.springframework.web.server.ResponseStatusException;

import java.io.IOException;
import java.util.List;
import java.util.stream.Collectors;

import static org.springframework.http.HttpStatus.BAD_REQUEST;
import static org.springframework.http.HttpStatus.NOT_FOUND;

@RestController
@Tag(name = "Übungsverwaltung")
public class ExerciseController {
    @Autowired
    private ExerciseMapper exerciseMapper;

    @Autowired
    private FileRefMapper fileRefMapper;

    @Autowired
    private ExerciseService exerciseService;

    @Autowired
    private PrincipalService principalService;

    @Autowired
    private ExerciseProgressResultMapper exerciseProgressResultMapper;

    @Autowired
    private ExerciseProgressService exerciseProgressService;

    @Autowired
    private TaskInstanceMapper taskInstanceMapper;

    @Autowired
    private TaskInstanceService taskInstanceService;

    @Autowired
    private TaskService taskService;

    @Autowired
    private CourseService courseService;

    @Autowired
    private PolicyEvaluator policy;

    @Autowired
    private FileResponseBuilder fileResponseBuilder;

    @GetMapping("/exercise/{id}")
    @Operation(summary = "Liefert eine Übung mit nur den Aufgabeninstanzen, die zu unarchivierten Aufgaben gehören")
    @ApiResponses(value = {
            @ApiResponse(responseCode = "404", description = ExerciseNotFoundException.REASON),
    })
    ExerciseDto getOne(@PathVariable Long id) {
        try {
            var exercise = exerciseService.findById(id);
            policy.checkAndRaise(CheckPolicyRequest.fromPrincipal(exercise, ExercisesAction.EXERCISE_VIEW));
            return exerciseMapper.map(exercise);
        } catch(ExerciseNotFoundException ex) {
            throw new ResponseStatusException(NOT_FOUND, ExerciseNotFoundException.REASON);
        }
    }

    @PostMapping("/exercise")
    @Operation(summary = "Erstellt eine Übung")
    @ApiResponses(value = {
            @ApiResponse(responseCode = "400", description = CourseNotFoundException.REASON),
    })
    ExerciseDto create(@RequestBody CreateExerciseDto dto) {
        try {
            var course = courseService.findById(dto.getCourseId());
            policy.checkAndRaise(CheckPolicyRequest.fromPrincipal(course, ExercisesAction.EXERCISE_CREATE));

            var exercise = exerciseService.create(course, dto);
            return exerciseMapper.map(exercise);
        } catch(CourseNotFoundException ex) {
            throw new ResponseStatusException(BAD_REQUEST, CourseNotFoundException.REASON);
        }
    }

    @PutMapping ("/exercise/{id}")
    @Operation(
            summary = "Aktualisiert eine Übung",
            description = "Wenn dabei eine Aufgabeninstanz gelöscht wird, zu der Lösungen existieren, wird ein spezielles Fehlerobjekt (ToDo) mit dem Status 400 zurückgegeben, welches angibt, welche Aufgabeninstanz betroffen ist und wie viele Lösung gelöscht würden. Wenn das Update dennoch erfolgen soll, muss die Anfrage mit dem ConfirmDangerousAction Header wiederholt werden."
    )
    @ApiResponses(value = {
            @ApiResponse(responseCode = "400", description = DangerousActionException.REASON),
            @ApiResponse(responseCode = "404", description = ExerciseNotFoundException.REASON),
    })
    ExerciseDto update(
            @PathVariable Long id,
            @RequestBody CreateExerciseDto dto,
            @Parameter(description = "Bestätigung der gefährlichen Aktion mit dem Wert true. Bei initialer Anfrage leer lassen und erst nach expliziter Bestätigung durch den Benutzer (über Dialog o.ä.) bei der wiederholten Anfrage diesen Header setzen.")
            @RequestHeader(value = "ConfirmDangerousAction", required = false) boolean confirmDangerousAction
    ) {
        try {
            var exercise = exerciseService.findById(id);
            policy.checkAndRaise(CheckPolicyRequest.fromPrincipal(exercise, ExercisesAction.EXERCISE_UPDATE));

            exercise = exerciseService.update(exercise, dto, confirmDangerousAction);
            return exerciseMapper.map(exercise);
        } catch(ExerciseNotFoundException ex) {
            throw new ResponseStatusException(NOT_FOUND, ExerciseNotFoundException.REASON);
        }
    }

    @GetMapping("/exercise")
    @Operation(summary = "Liefert alle Übungen zu einem Kurs")
    @ApiResponses(value = {
            @ApiResponse(responseCode = "400", description = CourseNotFoundException.REASON),
    })
    List<ExerciseDto> getExercises(@RequestParam("courseId") Long courseId) {
        try {
            var exercises = exerciseService.findAll(
                    new ExerciseFilter().courseId(courseId),
                    ExercisesAction.EXERCISE_VIEW
            );

            return exercises.stream()
                    .map(e -> exerciseMapper.map(e))
                    .collect(Collectors.toList());
        } catch(CourseNotFoundException ex) {
            throw new ResponseStatusException(BAD_REQUEST, CourseNotFoundException.REASON);
        }
    }

    @GetMapping("/exercise/progress")
    @Operation(summary = "Liefert Informationen zum Fortschritt (Zusammenfassung) eines Benutzers über alle Übungen eines Kurses hineweg")
    ExerciseProgressResultDto getProgressForUser(@RequestParam Long courseId, @RequestParam Long userId) {
        try {
            policy.checkAndRaise(CheckPolicyRequest.fromPrincipal(Pair.of(courseId, userId), CoursesAction.COURSE_PROGRESS_VIEW));

            var result = exerciseProgressService.generateFor(courseId, userId);
            return exerciseProgressResultMapper.map(result);
        } catch(CourseNotFoundException ex) {
            throw new ResponseStatusException(BAD_REQUEST, CourseNotFoundException.REASON);
        }
    }

    @PutMapping("/exercise/{id}/task/{taskId}/public")
    @Operation(summary = "Veröffentlicht (oder versteckt) die Musterlösung, die mit der Aufgabe verknüpft ist, im Rahmen der angegebenen Aufgabeninstanz")
    @ApiResponses(value = {
            @ApiResponse(responseCode = "404", description = TaskInstanceNotFoundException.REASON),
    })
    void publishSolution(@PathVariable Long id, @PathVariable Long taskId, @RequestBody UpdatePublicDto dto) {
        try {
            var taskInstance = taskInstanceService.findByExerciseAndTask(id, taskId);
            policy.checkAndRaise(CheckPolicyRequest.fromPrincipal(taskInstance.getExercise(), ExercisesAction.EXERCISE_UPDATE));

            taskInstanceService.publishSolution(taskInstance, dto.isPub());
        } catch(TaskInstanceNotFoundException ex) {
            throw new ResponseStatusException(NOT_FOUND, TaskInstanceNotFoundException.REASON);
        }
    }

    @GetMapping("/exercise/{id}/task/{taskId}")
    @Operation(summary = "Liefert eine Aufgabeninstanz, die eine konfigurierte Aufgabe in einem Übungsblatt mit zusätzlichen Informationen (Punktzahl etc.) enthält")
    @ApiResponses(value = {
            @ApiResponse(responseCode = "404", description = TaskInstanceNotFoundException.REASON),
    })
    TaskInstanceDetailDto getTaskInstance(@PathVariable Long id, @PathVariable Long taskId) {
        try {
            var taskInstance = taskInstanceService.findByExerciseAndTask(id, taskId);
            policy.checkAndRaise(CheckPolicyRequest.fromPrincipal(taskInstance.getExercise(), ExercisesAction.EXERCISE_VIEW));

            return taskInstanceMapper.mapDetail(taskInstance);
        } catch(TaskInstanceNotFoundException ex) {
            throw new ResponseStatusException(NOT_FOUND, TaskInstanceNotFoundException.REASON);
        }
    }

    @GetMapping("/exercise/{id}/task/{taskId}/asset")
    @Operation(
            summary = "Endpunkt /task/{id}/asset nur nochmal für Aufgabeninstanzen zwecks Berechtigungsprüfung"
    )
    @ApiResponses(value = {
            @ApiResponse(responseCode = "404", description = TaskNotFoundException.REASON + " oder " + AssetNotFoundException.REASON),
    })
    ResponseEntity<Resource> downloadAsset(@PathVariable Long id, @PathVariable Long taskId, @RequestParam("path") String path) throws IOException {
        try {
            var taskInstance = taskInstanceService.findByExerciseAndTask(id, taskId);
            policy.checkAndRaise(CheckPolicyRequest.fromPrincipal(taskInstance.getExercise(), ExercisesAction.EXERCISE_VIEW));

            var file = taskService.getTextAsset(taskInstance.getTask(), path);
            var resource = ResourceHelper.load(file);
            return fileResponseBuilder.build(resource, file.getName(), true);
        } catch(TaskNotFoundException ex) {
            throw new ResponseStatusException(NOT_FOUND, TaskNotFoundException.REASON);
        } catch(AssetNotFoundException ex) {
            throw new ResponseStatusException(NOT_FOUND, "%s: %s".formatted(AssetNotFoundException.REASON, ex.getPath(), id));
        }
    }

    @GetMapping("/exercise/{id}/task/{taskId}/files")
    @Operation(summary = "Für Bearbeitung der Aufgabe relevante Dateien, gefiltert nach Sichtbarkeit im Kontext des angemeldeten Nutzers. Lösungsdateien werden z.B. nicht Studenten angezeigt.")
    @ApiResponses(value = {
            @ApiResponse(responseCode = "400", description = TaskNotFoundException.REASON),
            @ApiResponse(responseCode = "404", description = TaskInstanceNotFoundException.REASON),
    })
    List<FileRefDto> getTaskInstanceFiles(@PathVariable Long id, @PathVariable Long taskId) {
        try {
            var taskInstance = taskInstanceService.findByExerciseAndTask(id, taskId);
            var files = taskInstanceService.getFiles(taskInstance);

            files = files.stream()
                    .filter(file -> policy.check(CheckPolicyRequest.fromPrincipal(Pair.of(file, taskInstance), ExercisesAction.TASK_FILE_VIEW)))
                    .toList();

            return files.stream().map(task -> fileRefMapper.map(task)).collect(Collectors.toList());
        } catch(TaskNotFoundException ex) {
            throw new ResponseStatusException(BAD_REQUEST, TaskNotFoundException.REASON);
        } catch(TaskInstanceNotFoundException ex) {
            throw new ResponseStatusException(NOT_FOUND, TaskInstanceNotFoundException.REASON);
        }
    }

    @GetMapping("/exercise/{id}/task/{taskId}/files/{uuid}/download")
    @Operation(summary = "Zugriff auf den Inhalt der Datei zu einer Aufgabeninstanz")
    @ApiResponses(value = {
            @ApiResponse(responseCode = "404", description = TaskInstanceNotFoundException.REASON + " oder " + FileNotFoundException.REASON),
    })
    ResponseEntity<Resource> downloadFile(@PathVariable Long id, @PathVariable Long taskId, @PathVariable String uuid) throws IOException {
        try {
            var taskInstance = taskInstanceService.findByExerciseAndTask(id, taskId);
            var fileRef = taskService.getFileRefById(taskInstance.getTask(), uuid);
            var file = taskService.getFile(taskInstance.getTask(), uuid);

            policy.checkAndRaise(CheckPolicyRequest.fromPrincipal(Pair.of(fileRef, taskInstance), ExercisesAction.TASK_FILE_VIEW));

            if (file.isDirectory()) {
                var zipFile = ZipHelper.zip(file, true);
                var resource = new UrlResource(zipFile.toURI());
                return fileResponseBuilder.build(resource, "%s.zip".formatted(file.getName()), true);
            } else {
                var resource = ResourceHelper.load(file);
                return fileResponseBuilder.build(resource, true);
            }
        } catch(FileNotFoundException e) {
            throw new ResponseStatusException(NOT_FOUND, FileNotFoundException.REASON);
        } catch(TaskInstanceNotFoundException ex) {
            throw new ResponseStatusException(NOT_FOUND, TaskInstanceNotFoundException.REASON);
        }
    }

    @DeleteMapping("/exercise/{id}")
    @Operation(summary = "Löscht eine Übung mit allen verknüpften Lösungen, Abgaben etc.")
    @ApiResponses(value = {
            @ApiResponse(responseCode = "404", description = ExerciseNotFoundException.REASON),
    })
    void delete(@PathVariable Long id) {
        try {
            var exercise = exerciseService.findById(id);
            policy.checkAndRaise(CheckPolicyRequest.fromPrincipal(exercise, ExercisesAction.EXERCISE_DELETE));

            exerciseService.delete(exercise);
        } catch(ExerciseNotFoundException ex) {
            throw new ResponseStatusException(NOT_FOUND, ExerciseNotFoundException.REASON);
        }
    }

    @GetMapping("/exercise/{id}/task")
    @Operation(summary = "Liefert die unarchivierten Aufgabeninstanzen, die zur Übung zugeordnet sind")
    @ApiResponses(value = {
            @ApiResponse(responseCode = "404", description = ExerciseNotFoundException.REASON),
    })
    List<TaskInstanceDetailDto> getTaskInstances(@PathVariable Long id) {
        try {
            var exercise = exerciseService.findById(id);
            policy.checkAndRaise(CheckPolicyRequest.fromPrincipal(exercise, ExercisesAction.EXERCISE_VIEW));

            var taskInstances = taskInstanceService.findAll(new TaskInstanceFilter().exerciseId(exercise.getId()));
            return taskInstances.stream().map(taskInstance -> taskInstanceMapper.mapDetail(taskInstance)).toList();
        } catch(ExerciseNotFoundException ex) {
            throw new ResponseStatusException(NOT_FOUND, ExerciseNotFoundException.REASON);
        }
    }
}
