package de.hsrm.sls.subato.exercises.exercise.progress;

import de.hsrm.sls.subato.auth.principal.PrincipalService;
import de.hsrm.sls.subato.auth.principal.SubatoPrincipal;
import de.hsrm.sls.subato.courses.course.Course;
import de.hsrm.sls.subato.courses.course.CourseService;
import de.hsrm.sls.subato.courses.membership.StaffGroup;
import de.hsrm.sls.subato.exercises.exercise.ExerciseFilter;
import de.hsrm.sls.subato.exercises.exercise.ExerciseService;
import de.hsrm.sls.subato.exercises.submission.SubmissionService;
import de.hsrm.sls.subato.exercises.submission.progress.SubmissionsProgress;
import de.hsrm.sls.subato.exercises.taskinstance.TaskInstanceFilter;
import de.hsrm.sls.subato.exercises.taskinstance.TaskInstanceService;
import de.hsrm.sls.subato.user.User;
import de.hsrm.sls.subato.user.UserService;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;

import java.util.ArrayList;
import java.util.Comparator;
import java.util.List;
import java.util.stream.Collectors;

@Service
public class ExerciseProgressService {
    @Autowired
    private SubmissionService submissionService;

    @Autowired
    private UserService userService;

    @Autowired
    private ExerciseService exerciseService;

    @Autowired
    private PrincipalService principalService;

    @Autowired
    private TaskInstanceService taskInstanceService;

    @Autowired
    private CourseService courseService;

    public ExerciseProgressResult generateFor(long courseId, long userId) {
        var course = courseService.findById(courseId);
        var user = userService.findById(userId);
        var currentUser = (SubatoPrincipal) principalService.getPrincipal();

        return generateFor(course, user, currentUser.hasGroup(course, StaffGroup.LECTURER) ||
                        currentUser.hasGroup(course, StaffGroup.TUTOR));
    }

    public ExerciseProgressResult generateFor(Course course, User user, boolean includeInvisibleExercises) {
        var exerciseProgresses = new ArrayList<ExerciseProgress>();
        var exercises = exerciseService.findAll(new ExerciseFilter().courseId(course.getId()));

        for(var exercise : exercises) {
            if(!includeInvisibleExercises && !exercise.isVisible()) {
                continue;
            }

            var exerciseProgress = new ExerciseProgress();
            exerciseProgress.setExercise(exercise);

            var taskInstances = taskInstanceService.findAll(new TaskInstanceFilter().exerciseId(exercise.getId()));
            exerciseProgress.setTaskInstances(taskInstances);

            var submission = submissionService.findByExerciseAndUser(exercise.getId(), user.getId());
            if(submission != null) {
                exerciseProgress.setSubmission(submission);

                var solutionsToConsider = submission.getSolution().stream().toList();
                var taskInstanceSolutions = solutionsToConsider.stream().filter(s -> s.getTaskInstance() != null).collect(Collectors.groupingBy(s -> s.getTaskInstance()));
                var latestSolutions = taskInstanceSolutions.entrySet().stream().map(entry -> entry.getValue().stream().findFirst().get()).collect(Collectors.toList());
                exerciseProgress.setSolutions(latestSolutions);

                var testsTotal = latestSolutions.stream().filter(s -> s.getEvalResult() != null).map(s -> s.getEvalResult().getTestsRun()).reduce(0, (a, b) -> a + b);
                var testsPassed = latestSolutions.stream().filter(s -> s.getEvalResult() != null).map(s -> s.getEvalResult().getTestsPassed()).reduce(0, (a, b) -> a + b);
                var leaks = latestSolutions.stream().filter(s -> s.getEvalResult() != null).map(s -> s.getEvalResult().getLeaks()).reduce(0, (a, b) -> a + b);
                var styleErrors = latestSolutions.stream().filter(s -> s.getEvalResult() != null).map(s -> s.getEvalResult().getStyleCheckResults().size()).reduce(0, (a, b) -> a + b);

                exerciseProgress.setTestsTotal(testsTotal);
                exerciseProgress.setTestsPassed(testsPassed);
                exerciseProgress.setLeaksTotal(leaks);
                exerciseProgress.setStyleErrorsTotal(styleErrors);
            }

            exerciseProgresses.add(exerciseProgress);
        }
        exerciseProgresses.sort(Comparator.comparing((ExerciseProgress exerciseProgress) -> exerciseProgress.getExercise().getId()).reversed());

        var submissions = exerciseProgresses.stream()
                .filter(p -> p.getSubmission() != null)
                .map(p -> p.getSubmission())
                .toList();
        var submissionsProgress = SubmissionsProgress.fromSubmissions(submissions);
        return new ExerciseProgressResult(exerciseProgresses, submissionsProgress);
    }
}
