package de.hsrm.sls.subato.modules.term;

import de.hsrm.sls.subato.auth.policy.CheckPolicyRequest;
import de.hsrm.sls.subato.auth.policy.PolicyEvaluator;
import de.hsrm.sls.subato.auth.policy.modules.ModulesAction;
import de.hsrm.sls.subato.shared.ExceptionReasons;
import io.swagger.v3.oas.annotations.Operation;
import io.swagger.v3.oas.annotations.responses.ApiResponse;
import io.swagger.v3.oas.annotations.responses.ApiResponses;
import io.swagger.v3.oas.annotations.tags.Tag;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.security.access.prepost.PreAuthorize;
import org.springframework.web.bind.annotation.*;
import org.springframework.web.server.ResponseStatusException;

import java.util.List;
import java.util.stream.Collectors;

import static org.springframework.http.HttpStatus.NOT_FOUND;

@RestController
@Tag(name = "Modulverwaltung")
public class TermController {
    @Autowired
    private TermService termService;

    @Autowired
    private TermMapper termMapper;

    @Autowired
    private PolicyEvaluator policy;

    @GetMapping("/term/{id}")
    @ApiResponses(value = {
            @ApiResponse(responseCode = "404", description = TermNotFoundException.REASON),
    })
    TermDto getOne(@PathVariable Long id) {
        try {
            var term = termService.findById(id);
            policy.checkAndRaise(CheckPolicyRequest.fromPrincipal(term, ModulesAction.TERM_VIEW));

            return termMapper.map(term);
        } catch(TermNotFoundException ex) {
            throw new ResponseStatusException(NOT_FOUND, TermNotFoundException.REASON);
        }
    }

    @GetMapping("/term")
    @Operation(summary = "Alle sichtbaren Semester im Kontext des angemeldeten Benutzers")
    List<TermDto> getAll() {
        var terms = termService.findAll();
        terms = terms.stream()
                .filter(term -> policy.check(CheckPolicyRequest.fromPrincipal(term, ModulesAction.TERM_VIEW)))
                .collect(Collectors.toList());
        return terms.stream().map(t -> termMapper.map(t)).collect(Collectors.toList());
    }

    @PostMapping("/term")
    @PreAuthorize("hasPermission(null, @ModulesAction.TERM_CREATE)")
    @ApiResponses(value = {
            @ApiResponse(responseCode = "400", description = ExceptionReasons.CONSTRAINT_VALIDATION),
    })
    TermDto create(@RequestBody CreateTermDto dto) {
        var term = termService.create(dto);
        return termMapper.map(term);
    }

    @PutMapping ("/term/{id}")
    @PreAuthorize("hasPermission(null, @ModulesAction.TERM_UPDATE)")
    @ApiResponses(value = {
            @ApiResponse(responseCode = "400", description = ExceptionReasons.CONSTRAINT_VALIDATION),
            @ApiResponse(responseCode = "404", description = TermNotFoundException.REASON),
    })
    TermDto update(@PathVariable Long id, @RequestBody CreateTermDto dto) {
        try {
            var term = termService.update(id, dto);
            return termMapper.map(term);
        } catch(TermNotFoundException ex) {
            throw new ResponseStatusException(NOT_FOUND, TermNotFoundException.REASON);
        }
    }
}
