package de.hsrm.sls.subato.shared.mime;

import de.hsrm.sls.subato.aor.core.CoreDataImportRunner;
import org.apache.tika.Tika;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.stereotype.Component;

import java.io.File;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * Responsible for mapping file names to mime types.
 *
 * Uses the Apache Tika library for reliable detection.
 * However, some files are detected as application/octet-stream by tika, so we need to map the exceptions here
 */
@Component
public class MimeDetector {
    /**
     * configurable mapping rules
     */
    private MimeProperties mimeProperties;

    Logger logger = LoggerFactory.getLogger(MimeDetector.class);

    public MimeDetector(MimeProperties mimeProperties) {
        this.mimeProperties = mimeProperties;

        logger.info("Configured mime mappings: %s".formatted(mimeProperties.getMime()));
    }

    private Map<String, String> mergeMappings(List<MimeMapping> globalMappings, List<MimeMapping> overrides) {
        var mimeMappings = new ArrayList<MimeMapping>();
        mimeMappings.addAll(globalMappings);
        mimeMappings.addAll(overrides);

        var d = new HashMap<String, String>();
        for(var mapping : mimeMappings) {
            d.put(mapping.getMap(), mapping.getAs());
        }

        return d;
    }

    /**
     * @return actual content type or application/octet-stream if it can't be guessed
     */
    public String getContentType(String fileName, List<MimeMapping> mimeMappingOverrides) {
        var mergedMappings = mergeMappings(mimeProperties.getMime(), mimeMappingOverrides);

        for(var entry : mergedMappings.entrySet()) {
            var map = entry.getKey();
            var as = entry.getValue();

            if(fileName.endsWith(map)) {
                return as;
            }
        }

        String contentType = new Tika().detect(fileName);
        return contentType;
    }

    public String getContentType(String fileName) {
        return getContentType(fileName, new ArrayList<>());
    }

    public String getContentType(File file, List<MimeMapping> mimeMappingOverrides) {
        return getContentType(file.getName(), mimeMappingOverrides);
    }
}
