package de.hsrm.sls.subato.solutions.solution;

import de.hsrm.sls.subato.auth.policy.CheckPolicyRequest;
import de.hsrm.sls.subato.auth.policy.PolicyEvaluator;
import de.hsrm.sls.subato.exercises.taskinstance.TaskInstance;
import de.hsrm.sls.subato.shared.fs.FileTree;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.domain.Page;
import org.springframework.data.domain.Pageable;
import org.springframework.data.jpa.domain.Specification;
import org.springframework.stereotype.Service;

import java.io.File;
import java.util.UUID;

/**
 * Responsible for CRUD of solutions
 */
@Service
public class SolutionService {
    @Autowired
    private SolutionRepository solutionRepository;

    @Autowired
    private SolutionContentRepository solutionContentRepository;

    @Autowired
    private PolicyEvaluator evaluator;


    /**
     * @throws SolutionNotFoundException
     */
    public Solution findById(long id) {
        return solutionRepository.findById(id).orElseThrow(() -> new SolutionNotFoundException(id));
    }

    /**
     * Finds all solutions matching the given filter
     * and sorts them by their submission date (desc)
     */
    public Page<Solution> findAll(Pageable page, SolutionFilter filter, String permission) {
        Specification<Solution> spec = filter.toSpecification();

        if(permission != null) {
            var authSpec = evaluator.checkAsSpec(CheckPolicyRequest.fromPrincipal(null, permission));
            spec = spec.and(authSpec);
        }

        var solutions = solutionRepository.findAll(spec, page);
        return solutions;
    }

    public Long getSolutionCountForTaskInstance(TaskInstance taskInstance) {
        var spec = new SolutionFilter()
                .exerciseId(taskInstance.getExercise().getId())
                .taskId(taskInstance.getTask().getId()).toSpecification();

        return solutionRepository.count(spec);
    }

    public void deleteByTaskInstance(TaskInstance taskInstance) {
        var spec = new SolutionFilter()
                .exerciseId(taskInstance.getExercise().getId())
                .taskId(taskInstance.getTask().getId()).toSpecification();

        var solutions = solutionRepository.findAll(spec);
        for(var s : solutions) {
            delete(s);
        }
    }

    public void delete(Solution solution) {
        solutionRepository.delete(solution);
        solutionContentRepository.delete(solution);
    }

    public File getFile(Solution solution, UUID uuid) {
        return solutionContentRepository.getFile(solution, uuid);
    }

    File getSolutionAsZip(Solution solution) {
        return solutionContentRepository.getSolutionAsZip(solution);
    }

    public FileTree getTree(Solution solution) {
        return solutionContentRepository.getTree(solution);
    }
}
