package de.hsrm.sls.subato.tasks.pool.sync;


import de.hsrm.sls.subato.tasks.imp.TaskImportResult;
import de.hsrm.sls.subato.tasks.imp.TaskImportService;
import de.hsrm.sls.subato.tasks.pool.TaskPool;
import de.hsrm.sls.subato.tasks.pool.TaskPoolRepository;
import de.hsrm.sls.subato.tasks.pool.sync.repository.RepositoryDoesNotExistException;
import de.hsrm.sls.subato.tasks.pool.sync.repository.RepositoryException;
import de.hsrm.sls.subato.tasks.pool.sync.repository.TaskPoolRepositoryManager;
import org.eclipse.jgit.api.Git;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

import java.time.LocalDateTime;

@Service
public class TaskPoolSyncService {
    @Autowired
    TaskImportService taskImportService;

    @Autowired
    TaskPoolRepository taskPoolRepository;

    @Autowired
    TaskPoolRepositoryManager repositoryManager;

    /**
     * checks if the repository for the given task pool has changed and if it has, reimports/archives the tasks in the repository
     * @param pool pool to sync
     * @param forceUpdate if true, the sync will be run - even if the repository did not change
     * @return null if no sync was performed (because there were no changes), otherwise the result of the sync
     */
    @Transactional
    public TaskImportResult syncPool(TaskPool pool, boolean forceUpdate) {
        if(!pool.isLinkedToRepository()) {
            throw new RepositoryDoesNotExistException();
        }

        try {
            Git repo;
            var repositoryChanged = false;
            if (!repositoryManager.repositoryExists(pool)) {
                repo = repositoryManager.cloneRepository(pool);
                repositoryChanged = true;
            } else {
                repo = repositoryManager.getRepository(pool);
                repositoryChanged = repositoryManager.refreshRepository(repo);
            }

            if (pool.getLastChecked() == null) {
                repositoryChanged = true;
            }

            var now = LocalDateTime.now();
            pool.setLastChecked(now);
            pool = taskPoolRepository.save(pool);

            if (repositoryChanged || forceUpdate) {
                var result = taskImportService.importTasks(pool, repo.getRepository().getDirectory().getParentFile());

                pool.setLastUpdated(now);
                pool = taskPoolRepository.save(pool);
                return result;
            }

            return null;
        } catch(Exception ex) {
            throw new TaskPoolSyncException(ex);
        }
    }
}
