package de.hsrm.sls.subato.tasks.task;

import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.core.type.TypeReference;
import com.fasterxml.jackson.databind.ObjectMapper;
import de.hsrm.sls.subato.shared.fs.FileTreeService;
import de.hsrm.sls.subato.shared.config.SubatoConfig;
import de.hsrm.sls.subato.shared.helper.FileHelper;
import de.hsrm.sls.subato.shared.helper.ZipHelper;
import de.hsrm.sls.subato.tasks.imp.ImportSpecRequest;
import de.hsrm.sls.subato.tasks.imp.TaskImportException;
import de.hsrm.sls.subato.tasks.stef.StefSpec;
import de.hsrm.sls.subato.tasks.stef.file.FileRef;
import de.hsrm.sls.subato.tasks.stef.file.FileType;
import de.hsrm.sls.subato.tasks.stef.parser.SpecParser;
import net.lingala.zip4j.model.ExcludeFileFilter;
import org.apache.commons.io.FileUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;
import org.springframework.util.FileSystemUtils;

import java.io.File;
import java.io.IOException;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.List;
import java.util.Optional;
import java.util.UUID;

@Service
public class TaskContentRepository {
    @Autowired
    private SubatoConfig subatoConfig;

    @Autowired
    private FileTreeService fileTreeService;

    @Autowired
    private SpecParser specParser;

    public Path getTaskPath(Task task) {
        return Paths.get(subatoConfig.getDataDir())
                .resolve("task")
                .resolve(String.valueOf(task.getId()));
    }

    public StefSpec getSpec(Task task) {
        var indexFile = getIndexFile(task);
        if(!indexFile.exists()) {
            var spec = specParser.parseSpec(getTaskPath(task).toString());
            createIndex(spec, task);
            return spec;
        }

        var objectMapper = new ObjectMapper();
        try {
            var specContent = Files.readString(indexFile.toPath(), StandardCharsets.UTF_8);
            var spec = objectMapper.readValue(specContent, StefSpec.class);
            return spec;
        } catch (IOException e) {
            throw new RuntimeException(e);
        }
    }

    public List<FileRef> getFileRefs(Task task) {
        return getSpec(task).getMeta().getFiles();
    }

    public Optional<FileRef> getFileRefById(Task task, UUID uuid) {
        var files = getFileRefs(task);
        return files.stream().filter(f -> f.getFile().getUuid().equals(uuid)).findFirst();
    }

    public File getSpecAsZip(Task task) {
        var root = getTaskPath(task);
        ExcludeFileFilter excludeFilter = file -> file.toPath().equals(getIndexFile(task).toPath()) || file.toPath().equals(getBundleDir(task).toPath());
        return ZipHelper.zip(root.toFile(), false, excludeFilter);
    }

    public File getFile(Task task, Path path) {
        var basePath = getTaskPath(task);
        var file = basePath.resolve(path).toFile();
        return file;
    }

    public File getFileByRef(Task task, FileRef ref) {
        return ref.getFile().getFile(getTaskPath(task));
    }

    public void delete(Task task) {
        var taskPath = getTaskPath(task);
        FileSystemUtils.deleteRecursively(taskPath.toFile());
    }

    /**
     * stores/replaces the spec, currently stored in the directory dir, for the given task in the subato2-data directory.
     *
     * when importing a spec for an existing task, a backup is created and rolled back, if the import fails
     */
    public void store(ImportSpecRequest request, File dir) {
        var taskPath = getTaskPath(request.task());
        var backupPath = Path.of(taskPath + ".old");

        try {
            if(taskPath.toFile().exists()) {
                FileUtils.copyDirectory(taskPath.toFile(), backupPath.toFile());
            }

            // delete dir for existing task if already exists
            delete(request.task());
        } catch (IOException e) {
            throw new RuntimeException(e);
        }

        try {
            FileUtils.copyDirectory(dir, taskPath.toFile());

            // try to build the zip bundle
            buildBundle(request);

            // create an index for performant access
            createIndex(request.spec(), request.task());
        } catch (Exception e) {
            try {
                FileUtils.copyDirectory(backupPath.toFile(), taskPath.toFile());
            } catch (IOException ex) {
                throw new RuntimeException(ex);
            }

            throw new RuntimeException(e);
        } finally {
            FileSystemUtils.deleteRecursively(backupPath.toFile());
        }
    }

    private String getBundleFileName(Task task) {
        return task.getIdName();
    }

    private File getBundleDir(Task task) {
        return getTaskPath(task).resolve("files").resolve("bundle").toFile();
    }

    private File getBundleFile(Task task) {
        return getBundleDir(task).toPath().resolve("%s.zip".formatted(getBundleFileName(task))).toFile();
    }

    /**
     * generates a zip file, bundling all public source files (excluding the solution) associated with the given task.
     * then stores the zip file in the file system
     *
     * useful for students wanting to import the task into their IDE without explicitly downloading each file manually
     *
     */
    private void buildBundle(ImportSpecRequest request) {
        var taskPath = getTaskPath(request.task());
        var spec = request.spec();

        var srcFiles = spec.getMeta().getFiles().stream()
                .filter(fileRef -> fileRef.isPublic())
                .filter(fileRef -> fileRef.getType() == FileType.TEMPLATE ||
                        fileRef.getType() == FileType.INCLUDE ||
                        fileRef.getType() == FileType.TEST)
                .toList();

        var miscFiles = spec.getMeta().getFiles().stream()
                .filter(fileRef -> fileRef.isPublic())
                .filter(fileRef -> fileRef.getType() == FileType.MISC)
                .toList();
        var targetFiles = new ArrayList<FileRef>();
        targetFiles.addAll(srcFiles);
        targetFiles.addAll(miscFiles);

        if(targetFiles.size() == 0) {
            return;
        }

        var bundleDirName = getBundleFileName(request.task());
        var tempDir = taskPath.resolve(bundleDirName).toFile();
        tempDir.mkdir();
        tempDir.deleteOnExit();

        try {
            for (var fileRef : targetFiles) {
                try {
                    var targetPath = srcFiles.contains(fileRef)
                            ? tempDir.toPath().resolve("src").resolve(Path.of(fileRef.getPath()))
                            : tempDir.toPath().resolve("resources").resolve(Path.of(fileRef.getPath()));
                    var targetDirectory = targetPath.getParent();
                    Files.createDirectories(targetDirectory);

                    var bytes = FileHelper.toBytes(fileRef.getFile().getFile(taskPath));
                    Files.write(targetPath, bytes);
                } catch (IOException e) {
                    throw new TaskImportException("bundle could not be built", e);
                }
            }

            var bundleDir = getBundleDir(request.task()).toPath();
            Files.createDirectories(bundleDir);

            var bundleFile = getBundleFile(request.task());
            var zipFile = ZipHelper.zip(tempDir, bundleFile, false);
            var item = fileTreeService.createTreeFrom(zipFile, getTaskPath(request.task()).relativize(bundleDir).toString(), false, true, new ArrayList<>());

            var taskFile = new FileRef();
            taskFile.setPath(bundleFile.getName());
            taskFile.setType(FileType.BUNDLE);
            taskFile.setPublic(true);
            taskFile.setDescription("Enthält alle benötigten Dateien");

            taskFile.setFile(item);
            spec.getMeta().getFiles().add(taskFile);
        } catch (IOException e) {
            throw new RuntimeException(e);
        } finally {
            FileSystemUtils.deleteRecursively(tempDir);
        }
    }

    private File getIndexFile(Task task) {
        var taskPath = getTaskPath(task);
        return taskPath.resolve(".stef.cache.json").toFile();
    }

    private void createIndex(StefSpec spec, Task task) {
        try {
            ObjectMapper objectMapper = new ObjectMapper();
            var serialized = objectMapper.writeValueAsString(spec);
            var indexFile = getIndexFile(task);
            Files.write(indexFile.toPath(), serialized.getBytes(StandardCharsets.UTF_8));
        } catch (JsonProcessingException e) {
            throw new RuntimeException(e);
        } catch (IOException e) {
            throw new RuntimeException(e);
        }
    }
}
